/****************************************************************
+---------------------------------------------------------------+
|                      M A G N E T R O N                        |
|                      -----------------                        |
|                                                               |
|               Written by Steven De Toni Oct-2007              |
|                                                               |
|                This source code is Public Domain              |
+---------------------------------------------------------------+
*****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <stdarg.h>
#include <memory.h>

#include "lib/SFont.h"
#include "objs.h"
#include "advanced_enemy.h"
#include "basic_enemy.h"
#include "dynamic_enemy.h"
#include "walker_objects.h"
#include "player.h"
#include "level_design.h"
#include "background_flash.h"
#include "menus.h"


#ifndef _XBOX
#ifdef _WIN32
#	include <windows.h>
#	ifdef _DEBUG
#		ifdef _MSC_VER
#			define _CRTDBG_MAP_ALLOC  // Use Visual C++'s memory checking functionality
#			include <crtdbg.h>
#		endif 
#	endif
#endif
#endif

// #define SHOW_FPS

// ----------------------------------------------------------------------------

SFont_Font*			PGameFontTitle       = NULL;
SFont_Font*			PGameFontInfoBonus[] = {NULL, NULL, NULL, NULL};
SDL_Surface*		PScreen              = NULL;
SDL_Joystick*		PGameJoyStick        = NULL;
SDL_Rect			GameScreenRect	     = {0, 0, 0, 0};
FPSmanager			GameFPS;
BackgroundFlash		GameFlasher;
SDLKey				GameKeyboard[512];
bool                GamePlaying = false;;
GameConfig          GameInfo =  
{
    false, 
		16, 
		1,
    {SDLK_LEFT, SDLK_UP, SDLK_RIGHT, SDLK_DOWN, SDLK_a, SDLK_w, SDLK_d, SDLK_s},
    {USRJOY_LEFT, USRJOY_UP, USRJOY_RIGHT, USRJOY_DOWN, USRJOY_BUT3, USRJOY_BUT4, USRJOY_BUT2, USRJOY_BUT1},
    {
        {GAMETITLE, 100000, 30},
        {GAMETITLE,  90000, 29},
        {GAMETITLE,  80000, 28},
        {GAMETITLE,  70000, 27},
        {GAMETITLE,  60000, 26},
        {GAMETITLE,  50000, 25},
        {GAMETITLE,  40000, 24},
        {GAMETITLE,  30000, 23},
        {GAMETITLE,  20000, 22},
        {GAMETITLE,  10000, 21}
    }
};

ParticalCollection* PExpObjs          = NULL;
SDLSpriteList*		PEnemyList        = NULL;
SDLSpriteList*		PFriendsList      = NULL;
SDLSpriteList*		PElectrodesList   = NULL;
SDLSpriteList*		PPlayerBulletList = NULL;

void SDL_ColPrintf (SDL_Surface* pS, SFont_Font* pF, Uint8 r, Uint8 g, Uint8 b, int colidx, int x, int y, char* pszFormat, ...)
{
    char buf[1024];
	va_list arglist;
	
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);
    
	Uint8 sR = pF->Surface->format->palette->colors[colidx].r;
	Uint8 sG = pF->Surface->format->palette->colors[colidx].g;
	Uint8 sB = pF->Surface->format->palette->colors[colidx].b;

	SDL_Color c1 = {r,g,b,0};
	SDL_SetColors (pF->Surface, &c1, colidx, 1);

	SFont_Write (pS, pF, x, y, buf);

	SDL_Color c2 = {sR,sG,sB,0};
	SDL_SetColors (pF->Surface, &c2, colidx, 1);
}


void SDL_Printf (SDL_Surface* pS, SFont_Font* pF, int x, int y, char* pszFormat, ...)
{
    char buf[1024];
	va_list arglist;
	
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);
    	
	SFont_Write (pS, pF, x, y, buf);
}

void SDL_Printf (int x, int y, char* pszFormat, ...)
{
    char buf[1024];
	va_list arglist;
	
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);
    	
	SFont_Write (PScreen, PGameFontTitle, x, y, buf);
}

void DbgPrintf (char* pszFormat, ...)
{
    char buf[1024];
	va_list arglist;
	
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);  

#ifndef  _XBOX
#	ifdef _WIN32 
	OutputDebugString (buf); 	
#	endif
#else
	FILE* f = fopen (_SPATH "debug.txt", "a+");
	if (f != NULL)
	{
		fprintf (f, "%s\n", buf);
		fclose (f);
	}
#endif
}

#ifndef _XBOX
#	ifdef _WIN32 
void ErrPrintf (char* pszFormat, ...)
{
    char buf[1024];
	va_list arglist;
	
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);   	

	MessageBox (NULL, buf, "Error", MB_ICONSTOP);
}
#	endif
#elif _XBOX  
#	define ErrPrintf DbgPrintf
#else 
#	define ErrPrintf printf
#endif


void GamePlaySrnLock (Player& player, bool clearScreen)
{
	static SDL_Event event;
	static bool joyBut[8] = {false,false,false, false,  // (left) left, up, right, down, 
                             false,false,false, false}; // (right) left, up, right, down, 

    while (SDL_PollEvent(&event))
    {
		switch (event.type)
		{
			case SDL_QUIT:
				GameKeyboard[SDLK_ESCAPE] = (SDLKey)true;
				break;

			case SDL_KEYDOWN:
				GameKeyboard[event.key.keysym.sym] = (SDLKey)true;
				break;

			case SDL_KEYUP:
				GameKeyboard[event.key.keysym.sym] = (SDLKey)false;
				break;

			case SDL_JOYBUTTONDOWN:				
			case SDL_JOYBUTTONUP:
				{
					if (event.jbutton.which != 0)
						break;

#ifdef _XBOX
                    if (event.jbutton.button == USRJOY_BUT10)
					{
						GameKeyboard[SDLK_ESCAPE] = (SDLKey) (event.jbutton.state == SDL_PRESSED ? true : false);
						break;
					}
					else if (event.jbutton.button == USRJOY_BUT9)
					{
						GameKeyboard[SDLK_p] = (SDLKey) (event.jbutton.state == SDL_PRESSED ? true : false);
						break;
					}
#else
                    if (event.jbutton.button == USRJOY_BUT8)
					{
						GameKeyboard[SDLK_ESCAPE] = (SDLKey) (event.jbutton.state == SDL_PRESSED ? true : false);
						break;
					}
					else if (event.jbutton.button == USRJOY_BUT7)
					{
						GameKeyboard[SDLK_p] = (SDLKey) (event.jbutton.state == SDL_PRESSED ? true : false);
						break;
					}
#endif
					// scan through joystick events, and translate these to key presses.
					for (int ujIdx = 0; ujIdx < 8; ujIdx++)
					{
						if (GameInfo.userJoy[ujIdx] == event.jbutton.button)
							GameKeyboard[GameInfo.userKeys[ujIdx]] = (SDLKey) (event.jbutton.state == SDL_PRESSED ? true : false);
					}					
				}
				break;
				
			// Translate joystick buttons into keyboard buttons
			case SDL_JOYHATMOTION:
                {
				    memset (joyBut, false, sizeof(joyBut));
				    if (event.jhat.value & SDL_HAT_LEFT)
					    joyBut[0] = true;
				    if (event.jhat.value & SDL_HAT_UP)
					    joyBut[1] = true;
				    if (event.jhat.value & SDL_HAT_RIGHT)
					    joyBut[2] = true;
				    if (event.jhat.value & SDL_HAT_DOWN)
					    joyBut[3] = true;
	    
				    // scan through joystick events, and translate these to key presses.
				    for (int ujIdx = 0; ujIdx < 8; ujIdx++)
				    {
					    switch (GameInfo.userJoy[ujIdx])
					    {
						    case USRJOY_LEFT:
							    GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[0];
							    break;
						    case USRJOY_UP:
							    GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[1];
							    break;
						    case USRJOY_RIGHT:
							    GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[2];
							    break;
						    case USRJOY_DOWN:
							    GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[3];
							    break;    
					    }
				    }				
                }
				break;

			case SDL_JOYAXISMOTION:
				{
//DbgPrintf ("Move %d:%d %d %d\r\n", event.jaxis.which, event.jaxis.axis,
//           SDL_JoystickGetAxis(PGameJoyStick, 5), SDL_JoystickGetAxis(PGameJoyStick, 6));
					if (event.jaxis.which != 0)
						break;

                    switch (event.jaxis.axis)
					{
						case 0: // x
                            if (event.jaxis.value < -512)
                            {
							    joyBut[0] = true;
							    joyBut[2] = false;
                            }
                            else if (event.jaxis.value > 512)
                            {
								joyBut[2] = true;
								joyBut[0] = false;
                            }
                            else
                            {
								joyBut[0] = false;
								joyBut[2] = false;
                            }
							break;

						case 1: // y
                            if (event.jaxis.value < -512)
                            {
								joyBut[1] = true;
								joyBut[3] = false;
                            }
                            else if (event.jaxis.value > 512)
                            {
								joyBut[3] = true;
								joyBut[1] = false;
                            }
                            else
                            {
								joyBut[1] = false;
								joyBut[3] = false;
                            }
							break;

						case 2: // x
                            if (event.jaxis.value < -512)
                            {
							    joyBut[4] = true;
							    joyBut[6] = false;
                            }
                            else if (event.jaxis.value > 512)
                            {
								joyBut[6] = true;
								joyBut[4] = false;
                            }
                            else
                            {
								joyBut[4] = false;
								joyBut[6] = false;
                            }
							break;
						case 3: // y
                            if (event.jaxis.value < -512)
                            {
								joyBut[5] = true;
								joyBut[7] = false;
                            }
                            else if (event.jaxis.value > 512)
                            {
								joyBut[7] = true;
								joyBut[5] = false;
                            }
                            else
                            {
								joyBut[5] = false;
								joyBut[7] = false;
                            }
							break;
                    }

					// scan through joystick events, and translate these to key presses.
					for (int ujIdx = 0; ujIdx < 8; ujIdx++)
					{						
						switch (event.jaxis.axis)
						{
							// translate left joy stick
							case 0:
							case 1:
								switch (GameInfo.userJoy[ujIdx])
								{
									case USRJOY_LEFT:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[0];
										break;
									case USRJOY_UP:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[1];
										break;
									case USRJOY_RIGHT:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[2];
										break;
									case USRJOY_DOWN:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[3];
										break;		
								}
								break;

							// translate right joy stick
							case 2:
							case 3:
								switch (GameInfo.userJoy[ujIdx])
								{
									case USRJOY_BUT3:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[4];
										break;
									case USRJOY_BUT4:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[5];
										break;
									case USRJOY_BUT2:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[6];
										break;
									case USRJOY_BUT1:
										GameKeyboard[GameInfo.userKeys[ujIdx]]  = (SDLKey)joyBut[7];
										break;                                   
								}
						}
					}				
				}
				break;
		}
    }

    // ----------------------------------------     
    
	if ((GameKeyboard[SDLK_p] != false) && (GamePlaying != false))
			DoPauseScreen (player);

#ifndef _XBOX
    if ( SDL_MUSTLOCK(PScreen) )
    {
        if ( SDL_LockSurface(PScreen) < 0 )
        {
            return;
        }
    }
#endif

	if (clearScreen != false)
		SDL_FillRect(PScreen, &GameScreenRect, GameFlasher.update());
}

void GamePlaySrnUnlock (Player& player, bool dispPlayerScore)
{
    GameFlasher.drawGameBorder (GamePlayArea);

static Uint32 lastTick	  = GameFPS.currentticks;
static Uint32 lastFrameNo = GameFPS.framecount;
static Uint32 fps		  = 0;	   

	if ((GameFPS.currentticks - lastTick) > 1000)
	{
		fps			= (GameFPS.totalframecount - lastFrameNo) - 1;
		lastFrameNo = GameFPS.totalframecount;
		lastTick    = GameFPS.currentticks;
	}

    if (dispPlayerScore != false)
	{
        SDL_Printf (25,  2, "SCORE: %010d", player.playersScore);
		SDL_Printf (320, 2, "LIVES: %02d",  player.playerLives);
		SDL_Printf (470, 2, "WAVE: %02d",   player.gameLevel+1);

#ifdef SHOW_FPS
		SDL_Printf (605, 2, ":%d",		    fps);
#endif
	}

#ifndef _XBOX
    if ( SDL_MUSTLOCK(PScreen) )
    {
        SDL_UnlockSurface(PScreen);
    }
#endif

	SDL_Flip (PScreen);    		
	SDL_framerateDelay (&GameFPS);		        
}

void CheckSpawnCreateOnDeath (SDLSprite* pNode, ObjectType bulletHitType)
{
	SDLSpriteList* pList;

    if (pNode->m_DeathRelease != 0)
    {
        SDLSprite* pNewObj = CreateNewObject ((ObjectType) pNode->m_DeathRelease, (LevelListType)pNode->m_DeathRelAssocLstType);
        if (pNewObj == NULL)
            return;

        pNewObj->m_PosX = pNode->m_PosX + (pNode->m_BlockWidth/2) - (pNewObj->m_BlockWidth/2);
        pNewObj->m_PosY = pNode->m_PosY + (pNode->m_BlockHeight/2) - (pNewObj->m_BlockHeight/2);
        
        // Add new obj to a list
        switch (pNode->m_DeathRelAssocLstType)
        {
            case LLT_ENEMY:
				pList = PEnemyList;
                break;

            case LLT_ELECTRODE:
                pList = PElectrodesList;
                break;

            case LLT_FRIENDS:
			default:
				pList = PFriendsList;
                break;
        }

        pNewObj->MaimedHit (bulletHitType);
        pList->AddSprite (pNewObj);
    }
}

bool PlayerGotHitTest (SDLSpriteList* pObjectTestList, SDLSprite*& pNode, Player& player)
{
    if (pNode->SpriteHitPixel(&player) != SDL_FALSE)
    {      
		player.m_HitPoints--;
		if (player.m_HitPoints <= 0)  // if player has run out of hit points...
		{					
			player.PlaySnd (SNDT_DEATH);
            player.gotHit = true;
		}

        return true;
    }

    return false;
}

int SmartBombOnObjects (SDLSpriteList* pObjectList, Player& player)
{
    SDLSprite* pNode;
    int        hit  = 0;

    for (pNode = pObjectList->m_pList; pNode != NULL;)
	{
        pNode->m_HitPoints--;
        if (pNode->m_HitPoints <= 0) 
        {
			CheckSpawnCreateOnDeath (pNode);

			player.add2Score (pNode->m_Score);
            PExpObjs->buildExplosionSplit (pNode, (PCExplodeType)(rand() % 5));            
            hit++;

			pObjectList->DelSprite(pNode); 
            continue;
        }

        if (pNode != NULL)
            pNode = pNode->m_Next;
    }

    return hit;
}

void KillFriendlySkullCrossbones (SDLSprite* pNode)
{
    BasicEnemy* pSkull = new BasicEnemy (BET_SKULLCROSS1);            
    int wDiff = pNode->m_BlockWidth - pSkull->m_BlockWidth;
    int hDiff = pNode->m_BlockHeight- pSkull->m_BlockHeight;

	pSkull->m_PosX       = pNode->m_PosX + wDiff/2;
	pSkull->m_PosY       = pNode->m_PosY + wDiff/2;
    pSkull->m_TimeToLive = 100;

    SNDPlayPan (SNDCHL_GAMEMISC, GameSNDResource[GRS_ELECTRICDEATH].pResSnd, 0, (int)pNode->m_PosX+(pNode->m_BlockWidth/2), (int)pNode->m_PosY+(pNode->m_BlockHeight/2));
    PEnemyList->AddSprite ((SDLSprite*)pSkull);
}

bool PlayerBulletHitObjectTest (SDLSpriteList*      pObjectTestList, 
							    SDLSprite*&         pNode, 
								Player&             player)
{
    SDLSprite* pBulletNode;

    for (pBulletNode = PPlayerBulletList->m_pList; pBulletNode != NULL;)	
	{
		// Check if sprite is hit by bullet
		if (pNode->SpriteHitPixel(pBulletNode) != SDL_FALSE)
		{		
    		pNode->m_HitPoints -= pBulletNode->m_HitPoints;

			// do some hit processing ...
            pNode->MaimedHit ((int)pBulletNode->m_Type);

			if (pNode->m_HitPoints <= 0)  // check if enemy can take a number of hits?
			{				
				// Create death scene...
                switch (pNode->m_Type)
				{
					case AET_BIG_BRIAN_BOSS:
					case AET_ORB_BIG:
					case AET_BOSS_BRAIN:
					case AET_BOSS1:				
						switch (pBulletNode->m_Type)
						{
							// virtical
							case PLY_BULLET_LL:
							case PLY_BULLET_RR:
								PExpObjs->buildExplosionSplit (pNode, PC_VERT, false, 0.1f, 130);
								break;

							// horizontal
							case PLY_BULLET_UU:
							case PLY_BULLET_DD:
								PExpObjs->buildExplosionSplit (pNode, PC_HORZ, false, 0.1f, 130);
								break;
									

							// diagonal left
							case PLY_BULLET_RU:
							case PLY_BULLET_LD:
								PExpObjs->buildExplosionSplit (pNode, PC_DIAGLEFT, false, 0.1f, 130);
								break;
								
							// diagonal right
							case PLY_BULLET_LU:
							case PLY_BULLET_RD:
								PExpObjs->buildExplosionSplit (pNode, PC_DIAGRIGHT, false, 0.1f, 130);
								break;
						}                        
						break;

					default:
						switch (pBulletNode->m_Type)
						{
							// virtical
							case PLY_BULLET_LL:
							case PLY_BULLET_RR:
								PExpObjs->buildExplosionSplit (pNode, PC_VERT);
								break;

							// horizontal
							case PLY_BULLET_UU:
							case PLY_BULLET_DD:
								PExpObjs->buildExplosionSplit (pNode, PC_HORZ);
								break;
									

							// diagonal left
							case PLY_BULLET_RU:
							case PLY_BULLET_LD:
								PExpObjs->buildExplosionSplit (pNode, PC_DIAGLEFT);
								break;
								
							// diagonal right
							case PLY_BULLET_LU:
							case PLY_BULLET_RD:
								PExpObjs->buildExplosionSplit (pNode, PC_DIAGRIGHT);
								break;
						}					
				}
				


                CheckSpawnCreateOnDeath (pNode, (ObjectType)pBulletNode->m_Type);

				player.add2Score (pNode->m_Score);// adjust players score
                pNode->PlaySnd (SNDT_DEATH);

                // if hit points on bullet was more than hit point on object, 
                // then leave object in play and adjust bullet hit points
                if (pNode->m_HitPoints == 0)
                {
        		    PPlayerBulletList->DelSprite(pBulletNode);
					pObjectTestList->DelSprite(pNode);
                }
                else
                {
					pBulletNode->m_HitPoints = -pNode->m_HitPoints;
					pObjectTestList->DelSprite(pNode);
					if (pNode == NULL)
						return true;
                    continue; // check if another enemy can be hit..                    
                }			

                return true;
			}
            else
            {                              

				switch (pBulletNode->m_Type)
				{
					case PLY_BULLET_DD:                    
					case PLY_BULLET_UU:
                    case PLY_BULLET_RR:
                    case PLY_BULLET_LL:
                    case PLY_BULLET_RU:
					case PLY_BULLET_LD:
					case PLY_BULLET_LU:
					case PLY_BULLET_RD:

                        GameBMResource[GR_PLAYER_BULLET1].pResObj->m_PosX = pBulletNode->m_PosX;
                        GameBMResource[GR_PLAYER_BULLET1].pResObj->m_PosY = pBulletNode->m_PosY;
                        GameBMResource[GR_PLAYER_BULLET1].pResObj->SetYFrameBank (pBulletNode->GetYFrameBank());

						PExpObjs->buildExplosionSplit (GameBMResource[GR_PLAYER_BULLET1].pResObj, PC_DIAGLEFT);
                        PExpObjs->buildExplosionSplit (GameBMResource[GR_PLAYER_BULLET1].pResObj, PC_DIAGRIGHT);

                        GameBMResource[GR_PLAYER_BULLET1].pResObj->SetYFrameBank (0);
						break;
						
				}	

                // remove the bullet from the screen                
                pNode->PlaySnd (SNDT_HIT);
        		PPlayerBulletList->DelSprite(pBulletNode); 
                continue;                
            }
            
		}

        if (pBulletNode != NULL)
            pBulletNode = pBulletNode->m_Next;
	}

    return false;
}

bool ElectrodeHitFriendsList (SDLSprite*& pElectrodeNode)
{
    SDLSprite* pNode;

    for (pNode = PFriendsList->m_pList; pNode != NULL;)
	{
        // don't blow up POW type objects!
        if (pNode->m_IsBullet == SDL_FALSE)
		{
			// Check if sprite is hit by bullet
			if (pElectrodeNode->SpriteHitPixel(pNode) != SDL_FALSE)
			{		
				pNode->m_HitPoints--;
				if (pNode->m_HitPoints <= 0) 
				{				                          
					KillFriendlySkullCrossbones (pNode);

					// remove the friendly from the screen
        			PFriendsList->DelSprite(pNode);
					return true;
				}
			}
		}

        if (pNode != NULL)
            pNode = pNode->m_Next;
	}
    return false;
}

bool ElectrodeHitElectrodesList (SDLSprite*& pEnemyNode)
{
    SDLSprite* pNode;

	for (pNode = PElectrodesList->m_pList; pNode != NULL;)
	{
		// Check if sprite is hit by bullet
		if ((pEnemyNode->SpriteHitPixel(pNode) != SDL_FALSE) && (pNode != pEnemyNode))
		{		
			pNode->m_HitPoints--;

			// do some hit processing ...
            pNode->MaimedHit (-1);

			if (pNode->m_HitPoints <= 0) 
			{				          
				pNode->PlaySnd (SNDT_HIT);
                PExpObjs->buildExplosionSplit (pNode, PC_EXPOUT, false, 0.2f);

                // remove the friendly from the screen
        		PElectrodesList->DelSprite(pNode);
                return true;
            }
		}

        if (pNode != NULL)
            pNode = pNode->m_Next;
	}

    return false;
}


int ElectrodesUpdateList (Player& player)
{
	SDLSprite* pNode;	
    int        hit  = 0;

	// Loop the list and update the sprites
	for (pNode = PElectrodesList->m_pList; pNode != NULL;)
	{
        // Move sprites in this list
		pNode->Move (&player);

		// removed expired objects
		if (pNode->m_TimeToLive > 0)
		{
			pNode->m_TimeToLive--;
			if (pNode->m_TimeToLive == 0)
			{
				PElectrodesList->DelSprite(pNode); 
                continue;
			}
		}

        if (PlayerBulletHitObjectTest (PElectrodesList, pNode, player) != false)
        { 
            hit++;            
            continue;
        }

        if (ElectrodeHitFriendsList (pNode) != false)
        {
            hit++;            
            continue;
        }

        if (ElectrodeHitElectrodesList (pNode) != false)
        {
            hit++;            
            continue;
        }

		// Hit test against player.
        if (PlayerGotHitTest (PElectrodesList, pNode, player) != false)
		{
            hit++; 
			if (pNode == NULL)
				continue;
		}
        
        pNode->Draw(PScreen); 

        if (pNode != NULL)
            pNode = pNode->m_Next;
    }

    return hit;
}

void PlayerBulletUpdateList (void)
{
    static unsigned int frameNo = 0;
    SDLSprite*          pNode;

    frameNo++;

    // Loop the list and update the sprites
    for (pNode = PPlayerBulletList->m_pList; pNode != NULL;)
    {
        pNode->m_PosX += (pNode->m_VelX);
        pNode->m_PosY += (pNode->m_VelY);

        pNode->SetFrame(GameFPS.currentticks % pNode->m_NbTotFrame);

        // Remove objects from screen/list
        if (((pNode->m_PosX > PScreen->w) || (pNode->m_PosX < 0)) ||
           ((pNode->m_PosY > PScreen->h) || (pNode->m_PosY < 0)))
        {
            PPlayerBulletList->DelSprite(pNode);
            continue;
        }
        PPlayerBulletList->Draw(PScreen, 0, 0);

        if (pNode != NULL)
            pNode = pNode->m_Next;        
    }   
}

// return number hits
int FriendsUpdateList (Player& player)
{
	SDLSprite* pNode;	
    int        hit  = 0;
	
	// Loop the list and update the sprites
	for (pNode = PFriendsList->m_pList; pNode != NULL;)
	{    
		// removed expired objects
		if (pNode->m_TimeToLive > 0)
		{
			pNode->m_TimeToLive--;
			if (pNode->m_TimeToLive == 0)
			{
		        switch (pNode->m_Type)
				{
					case DET_BONUS:
						PExpObjs->buildExplosionSplit (pNode, PC_EXPOUT, false, 0.1f, 100, 1);
						break;
				}

				PFriendsList->DelSprite(pNode);	
                continue;
			}
		}

        // Move sprites in this list
		if (pNode->m_Type == DET_BONUS)
			goto nodeDraw;

		pNode->Move (&player);
			
		// Hit test against player.
        if (pNode->SpriteHitPixel(&player) != SDL_FALSE)
        {      
			pNode->PlaySnd (SNDT_HIT);

            switch (pNode->m_Type)
            {
				case WKO_POWATTRACT:
					{
			            player.setPlayerPOW (PPOW_ATTRACT);

			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, "-= Magnetite =-");              
			            PFriendsList->AddSprite ((SDLSprite*)pNew);
			            pNew->m_PosX = pNode->m_PosX -  (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);	
					}
					break;

                case WKO_POWSHOCKBLASTER:
                    {
						player.setPlayerPOW (PPOW_SHOCKBLASTER);

			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, "-= Shock Blaster =-");              
			            PFriendsList->AddSprite ((SDLSprite*)pNew);
			            pNew->m_PosX = pNode->m_PosX -  (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);
                    }
                    break;

                case WKO_POWSMARTBOMB:
                    {
			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, "-= Smart Bomb =-");              
			            PFriendsList->AddSprite ((SDLSprite*)pNew);
			            pNew->m_PosX = pNode->m_PosX -  (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);

	    	            SmartBombOnObjects (PEnemyList, player);         
	                    SmartBombOnObjects (PElectrodesList, player);                                            
                    }
                    break;

                case WKO_POW3FIRE:
                    {
			            player.setPlayerPOW (PPOW_3WAYFIRE);

			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, "-= 3 Way Fire =-");              
			            PFriendsList->AddSprite ((SDLSprite*)pNew);
			            pNew->m_PosX = pNode->m_PosX -  (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);
                    }
                    break;

                case WKO_POWFASTERFIRE:
                    {
			            player.setPlayerPOW (PPOW_FASTERFIRE);

			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, "-= Faster Fire =-");              
			            PFriendsList->AddSprite ((SDLSprite*)pNew);
			            pNew->m_PosX = pNode->m_PosX -  (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);
                    }
                    break;

                case WKO_POWEXTRALIFE:
                    {
			            player.playerLives ++;

			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, "-= Extra Life =-");              
			            PFriendsList->AddSprite ((SDLSprite*)pNew);
			            pNew->m_PosX = pNode->m_PosX -  (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);
                    }
                    break;

                default:
                    {
			            // Display bonus tag on screen.
			            char buff[16];

			            // update player bonus multiplier
			            player.incrementBonusMultiplier ();
                        player.setPlayerPOW(PPOW_HOTBULLETS);

			            sprintf (buff, "%d", pNode->m_Score * player.bonusMultipler);                        
			            DynamicEnemy* pNew = new DynamicEnemy(DET_BONUS, buff); 
			            PFriendsList->AddSprite ((SDLSprite*)pNew);

                        // centre text
			            pNew->m_PosX = pNode->m_PosX - (pNew->m_BlockWidth/2);
			            pNew->m_PosY = pNode->m_PosY;
			            player.add2Score (pNode->m_Score);
                    }
            }

            hit++;
			PFriendsList->DelSprite(pNode); 
            continue;					
        }
nodeDraw:
        pNode->Draw(PScreen); 

        if (pNode != NULL)
            pNode = pNode->m_Next;
	}

    return hit;
}

bool EnemyHitElectrodesList (SDLSprite*& pEnemyNode)
{
    SDLSprite* pNode;

	for (pNode = PElectrodesList->m_pList; pNode != NULL;)
	{
		// Check if sprite is hit by bullet
		if (pEnemyNode->SpriteHitPixel(pNode) != SDL_FALSE)
		{	
            pNode->m_HitPoints--;            

			// do some hit processing ...
            pNode->MaimedHit (-1);
            
			if (pNode->m_HitPoints <= 0) 
			{				          
				pNode->PlaySnd (SNDT_HIT);
                PExpObjs->buildExplosionSplit (pNode, PC_EXPOUT, false, 0.2f);

                // remove the friendly from the screen
        		PElectrodesList->DelSprite(pNode);
                return true;
            }
		}

        if (pNode != NULL)
            pNode = pNode->m_Next;
	}

    return false;
}


// return number enemys hit
int EnemyUpdateList (Player& player)
{
	SDLSprite* pNode;
    int        hit  = 0;
	
	// Loop the list and update the sprites
	for (pNode = PEnemyList->m_pList; pNode != NULL;)
	{    
        // Move sprites in this list
		pNode->Move(&player);

		// removed expired objects
		if (pNode->m_TimeToLive > 0)
		{
			pNode->m_TimeToLive--;
			if (pNode->m_TimeToLive == 0)
			{
                switch (pNode->m_Type)
                {
                    case BET_SKULLCROSS1:
                        PExpObjs->buildExplosionSplit (pNode, PC_EXPOUT, false, 0.1f, 100, PARTBLKSIZE);
                        break;
                }
                PEnemyList->DelSprite(pNode);
				continue;
			}			
		}

        if (PlayerBulletHitObjectTest (PEnemyList, pNode, player) != false)
        {            
			if (pNode == NULL)
				break;
            hit++;            
        }

        // Check if enemy has hit and electrode, if so, explode electrode!
		if (pNode->m_IsBullet == SDL_FALSE) 
		{
			if (EnemyHitElectrodesList (pNode) != false)
			{
				hit++;            
				continue;
			}
		}

		// Test against player.
        if (PlayerGotHitTest (PEnemyList, pNode, player) != false)
        {
			if (pNode == NULL)
				break;
            hit++;
        }

        pNode->Draw(PScreen); 

        if (pNode != NULL)
            pNode = pNode->m_Next;
	}
    return hit;
}

bool LoadBitmapResources (void)
{
    int idx, fidx;

	for (idx = 0; GameBMResource[idx].score != -314159; idx++)
	{
        if (GameBMResource[idx].filename[0] != '\0')
        {
		    if ((GameBMResource[idx].pResObj = new SDLSprite (GameBMResource[idx].filename, GameBMResource[idx].frameWidth, GameBMResource[idx].frameHeight, GameBMResource[idx].frameNum)) == NULL)
				return false;

			if (GameBMResource[idx].pResObj->m_Surface == NULL)
				return false;

		    GameBMResource[idx].pResObj->SetColorKey (GameBMResource[idx].ckr, GameBMResource[idx].ckg, GameBMResource[idx].ckb);
        }
	}

	// Load and Prepare fonts
    SDL_Surface* f = IMG_Load (_SPATH "fonts" _SEP "sfont_game_blue_title.png");

    if (f == NULL)
	    goto error;

	if ((PGameFontTitle = SFont_InitFont(SDL_DisplayFormat(f))) == NULL)
        goto error;

    SDL_FreeSurface(f);

	// Load game colour changing fonts...
    {
	    char* pFilenames[] = {_SPATH "fonts" _SEP "sfont_small_game_red.bmp"};

	    for (fidx = 0; fidx < 1; fidx++)
	    {		
            if ((f = SDL_LoadBMP (pFilenames[fidx])) == NULL)
                goto error;
         
		    if ((PGameFontInfoBonus[fidx] = SFont_InitFont(f)) == NULL)
                goto error;
	    }
    }

	return true;
error:

    fprintf(stderr, "An error occured while loading the game fonts.");
    return false;
}

void FreeBitmapResources (void)
{
	for (int idx = 0; GameBMResource[idx].score != -314159; idx++)
	{
		if (GameBMResource[idx].pResObj != NULL)
		{
			delete GameBMResource[idx].pResObj; 
			GameBMResource[idx].pResObj = NULL;
		}
	}

	// Free fonts ... 
	SFont_FreeFont (PGameFontTitle);
	for (int fidx = 0; fidx < 1; fidx++)
		SFont_FreeFont (PGameFontInfoBonus[fidx]);
}

bool LoadSoundResources (void)
{
	for (int idx = 0; GameSNDResource[idx].filename[0] != '\0'; idx++)
	{
		if ((GameSNDResource[idx].pResSnd = Mix_LoadWAV(GameSNDResource[idx].filename)) == NULL)
			return false;	
	}

	return true;
}

void FreeSoundResources (void)
{
	for (int idx = 0; GameSNDResource[idx].filename[0] != '\0'; idx++)
	{
		if (GameSNDResource[idx].pResSnd != NULL)
			Mix_FreeChunk(GameSNDResource[idx].pResSnd);
	}
}

void SetGameTitle (void)
{
    SDL_WM_SetCaption (GAMETITLE " - " GAMETITLEPHRASE, GAMETITLE);   

    SDL_Surface* pBitmap = SDL_CreateRGBSurface (SDL_HWSURFACE, 32, 32, PScreen->format->BitsPerPixel, PScreen->format->Rmask, PScreen->format->Gmask, PScreen->format->Bmask, PScreen->format->Amask);                    

    SDL_Rect dr = {16-(GameBMResource[GR_PLAYER].frameWidth/2), 16 - (GameBMResource[GR_PLAYER].frameHeight/2), GameBMResource[GR_PLAYER].frameWidth, GameBMResource[GR_PLAYER].frameHeight};
    SDL_Rect sr = {0, 0, GameBMResource[GR_PLAYER].frameWidth, GameBMResource[GR_PLAYER].frameHeight};
  
    SDL_FillRect (pBitmap, &dr, 0x00000000); 
    SDL_BlitSurface (GameBMResource[GR_PLAYER].pResObj->m_Surface, &sr, pBitmap, &dr);

    SDL_WM_SetIcon (pBitmap, NULL);
    SDL_FreeSurface (pBitmap);
}

int GamePlayLoop (Player& player)
{
    LevelGoalType gameLevelGoal = LGT_KILLENEMY;

	player.gameLevel	= 0;
	player.playersScore = 0;
    GamePlaying			= true;

	// Setup skill level...
	switch (GameInfo.gameLevelAdj)
	{
		case 3:
			GameLevelAdj	   = 0.7f;
			player.playerLives = 10;
			break;

		case 2:
			GameLevelAdj	   = 1.0f;
			player.playerLives = 10;
			break;

		case 0:
			GameLevelAdj	   = 2.0f;
			player.playerLives = 3;
			break;

		case 1:
		default:
			GameLevelAdj	   = 1.0f;
			player.playerLives = 3;
	}

//*** DEBUG ***
//player.playerLives = 100;
//player.playersScore = 0;
//player.gameLevel    = 28;
//GameLevelAdj    += (GameLevelAdj * 1.2f);

    LoadLevel    (player.gameLevel % LEVELS_MAXLEVEL, player, gameLevelGoal);
    DoLevelStart (player);

    // -------------------------------------------------------------	 	

    while (GameKeyboard[SDLK_ESCAPE] == false) 
    {        
		GamePlaySrnLock (player);		

		player.processKeyboard ();

		// -=================================================================-
		// Draw Play Area:
		// -=================================================================-			

		ElectrodesUpdateList  (player);
		FriendsUpdateList     (player);
		EnemyUpdateList	      (player);                
		PlayerBulletUpdateList();

        player.Draw(PScreen);

        PExpObjs->moveShowAll (PScreen);


        // Player Died? 
        if (player.gotHit != false)
        {                        
            player.gotHit = false;
            player.playerLives--;  
            
            DoLevelDeath  (player);
            if (player.playerLives <= 0)
            {
                DoGameOver (player);
                break;
            }

            DoLevelStart (player);            
        }

        // Do next Level...
        if ((PEnemyList->m_nSprites <= 0) && (PExpObjs->listNo <= 0))
        {			
            player.gameLevel++;
			if ((player.gameLevel % (LEVELS_MAXLEVEL+1)) == (LEVELS_MAXLEVEL))
			{
                DisplayEndScreen (player);

				GameLevelAdj    += (GameLevelAdj * 1.2f);
                player.gameLevel++;
			}
			else
			{	
				DoLevelWarp  (player);   
				LoadLevel    ((player.gameLevel % LEVELS_MAXLEVEL), player, gameLevelGoal);
                DoLevelStart (player);
			}
        }
        
		// -=================================================================-

		GamePlaySrnUnlock (player);
    }

    GamePlaying = false;
    return 0;
}


#ifdef _XBOX
	int main(int argc, char* argv[])
#elif _WIN32
	int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
#else
	int main(int argc, char* argv[])
#endif
{

#ifndef _XBOX     
#	ifdef _WIN32
#		ifdef  _DEBUG 
#			ifdef _MSC_VER
			//	_crtBreakAlloc = 7838;
			_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF); 
#			endif
#		endif 
#	endif
#endif

    // ---------------------------------------------------------------------

    int rtnCode;

    ReadConfigFile ();

    rtnCode = 0;
    
Restart:

	if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0)
    {
        ErrPrintf("Unable to init SDL: %s\n", SDL_GetError());
        exit(1);
    }

	atexit (SDL_Quit);

    PScreen = SDL_SetVideoMode(640, 480, GameInfo.bitDepth, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_SRCALPHA | (GameInfo.fullScreen != false ? SDL_FULLSCREEN : 0));

    if (PScreen == NULL)
    {
        ErrPrintf("Unable to set vide mode 640 x 480 x %d bits : %s\n", GameInfo.bitDepth, SDL_GetError());
        exit(1);
    }

#ifdef _XBOX
	XBOX_SetScreenPos ();
#endif

    if ((PGameJoyStick = SDL_JoystickOpen (0)) != NULL)
	{
		SDL_JoystickEventState (SDL_ENABLE);
	}

	if (Mix_OpenAudio(44100, MIX_DEFAULT_FORMAT, 2, 1024)==-1) 
	{
	    ErrPrintf("Mix_OpenAudio: %s\n", Mix_GetError());
	    exit(2);
	}

	Mix_AllocateChannels (16);
    SDL_ShowCursor (SDL_DISABLE);

    // -------------------------------------------------------------

	if (LoadBitmapResources () == false)
	{
		ErrPrintf("Failed to load all bitmap resources.");
		exit(1);
	}

	if (LoadSoundResources () == false)
	{
		ErrPrintf("Failed to load all sound resources.");
		exit(1);
	}

    SetGameTitle ();

	srand(SDL_GetTicks());
    SQRTIntInit         ();
	BuildCosSinTable	();
	SDL_initFramerate   (&GameFPS);
	SDL_setFramerate    (&GameFPS, 60);    
   		
    Player              player;

	PExpObjs          = new ParticalCollection;
	PEnemyList        = new SDLSpriteList;	
	PFriendsList      = new SDLSpriteList;	
	PElectrodesList   = new SDLSpriteList;
	PPlayerBulletList = new SDLSpriteList;
    GamePlaying       = false;

	memset (GameKeyboard, false, sizeof(GameKeyboard));
		
	GameScreenRect.w = PScreen->w;
	GameScreenRect.h = PScreen->h;	

    while ((rtnCode = DisplayMainMenu (player)) == 0)
    {        
        GamePlayLoop (player);
    }

    if (rtnCode != 2)
        WriteConfigFile ();

	// Free memory ...
	FreeBitmapResources ();
	FreeSoundResources  ();

	Mix_CloseAudio();

	if (PGameJoyStick != NULL)
	    SDL_JoystickClose (PGameJoyStick);

	delete PPlayerBulletList;
	delete PElectrodesList;
	delete PEnemyList;
	delete PFriendsList;
    delete PExpObjs;

    SDL_Quit();

    if (rtnCode == 2) // Reset screen settings, they have changed.
    {
#ifdef _XBOX
        return 0;
#else
        goto Restart;
#endif    
    }

	return 0;
}

