/*

 SDL_framerate: framerate manager

 LGPL (c) A. Schiffler
 
 */

#include "SDL_framerate.h"

/* 
   Initialize the framerate manager
*/

void SDL_initFramerate(FPSmanager * manager)
{
    /*
     * Store some sane values 
     */
	manager->currentticks   = SDL_GetTicks();
	manager->totalframecount= 0;
    manager->framecount		= 0;
    manager->rate			= FPS_DEFAULT;
    manager->rateticks		= (1000.0f / (float) FPS_DEFAULT);
    manager->lastticks		= manager->currentticks;
}

/* 
   Set the framerate in Hz 
*/

int SDL_setFramerate(FPSmanager * manager, int rate)
{
    if ((rate >= FPS_LOWER_LIMIT) && (rate <= FPS_UPPER_LIMIT)) 
    {
	    manager->framecount = 0;
	    manager->rate = rate;
	    manager->rateticks = (1000.0f / (float) rate);
	    return (0);
    } 
    else 
    {
	    return (-1);
    }
}

/* 
  Return the current target framerate in Hz 
*/

int SDL_getFramerate(FPSmanager * manager)
{
    if (manager == NULL) 
    {
	    return (-1);
    } 
    else 
    {
	    return (manager->rate);
    }
}

/* 
  Delay execution to maintain a constant framerate. Calculate fps.
*/

void SDL_framerateDelay(FPSmanager * manager)
{
    Uint32 target_ticks;
    Uint32 the_delay;

    /*
     * Next frame 
     */
    manager->framecount++;
	manager->totalframecount++;

    /*
     * Get/calc ticks 
     */
    manager->currentticks = SDL_GetTicks();
    target_ticks  = manager->lastticks + (Uint32) ((float) manager->framecount * manager->rateticks);

    if (manager->currentticks <= target_ticks) 
    {
	    the_delay = target_ticks - manager->currentticks;
	    SDL_Delay(the_delay);
    } 
    else 
    {
	    manager->framecount = 0;
	    manager->lastticks  = manager->currentticks;
    }
}

